package net.hostsharing.hsadminng.rbac.rbacdef;

import java.util.function.Consumer;

import lombok.EqualsAndHashCode;
import lombok.Getter;
import net.hostsharing.hsadminng.hs.office.relationship.HsOfficeRelationshipEntity;
import net.hostsharing.hsadminng.rbac.rbacobject.RbacObject;

import jakarta.validation.constraints.NotNull;
import java.lang.reflect.InvocationTargetException;
import java.util.*;

import static net.hostsharing.hsadminng.rbac.rbacdef.RbacView.RbacUserReference.UserRole.CREATOR;
import static org.apache.commons.lang3.StringUtils.uncapitalize;

@Getter
public class RbacView {

    public static final String GLOBAL = "global";

    private final EntityAlias rootEntityAlias;

    private final Set<RbacUserReference> userDefs = new LinkedHashSet<>();
    private final Set<RbacRoleDefinition> roleDefs = new LinkedHashSet<>();
    private final Set<RbacPermissionDefinition> permDefs = new LinkedHashSet<>();
    private final Map<String, EntityAlias> entityAliases = new HashMap<>() {

        @Override
        public EntityAlias put(final String key, final EntityAlias value) {
            if ( containsKey(key) ) {
                throw new IllegalArgumentException("duplicate entityAlias: " + key);
            }
            return super.put(key, value);
        }
    };
    private final Set<String> updatableColumns = new LinkedHashSet<>();
    private final Set<RbacGrantDefinition> grantDefs = new LinkedHashSet<>();

    private SQL identityViewSqlQuery;
    private EntityAlias rootEntityAliasProxy;
    private RbacRoleDefinition previousRoleDef;

    public static <E extends RbacObject> RbacView rbacViewFor(final String alias, final Class<E> entityClass) {
        return new RbacView(alias, entityClass);
    }

    RbacView(final String alias, final Class<? extends RbacObject> entityClass) {
        rootEntityAlias = new EntityAlias(alias, entityClass);
        entityAliases.put(alias, rootEntityAlias);
        new RbacUserReference(CREATOR);
        entityAliases.put("global", new EntityAlias("global"));
    }

    public RbacView withUpdatableColumns(final String... columnNames) {
        Collections.addAll(updatableColumns, columnNames);
        return this;
    }

    public RbacView withIdentityView(final SQL sqlExpression) {
        this.identityViewSqlQuery = sqlExpression;
        return this;
    }

    public RbacView createRole(final Role role, final Consumer<RbacRoleDefinition> with) {
        final RbacRoleDefinition newRoleDef = findRbacRole(rootEntityAlias, role).toCreate();
        with.accept(newRoleDef);
        previousRoleDef = newRoleDef;
        return this;
    }

    public RbacView createSubRole(final Role role) {
        final RbacRoleDefinition newRoleDef = findRbacRole(rootEntityAlias, role).toCreate();
        findOrCreateGrantDef(newRoleDef, previousRoleDef).toCreate();
        previousRoleDef = newRoleDef;
        return this;
    }

    public RbacView createSubRole(final Role role, final Consumer<RbacRoleDefinition> with) {
        final RbacRoleDefinition newRoleDef = findRbacRole(rootEntityAlias, role).toCreate();
        findOrCreateGrantDef(newRoleDef, previousRoleDef).toCreate();
        with.accept(newRoleDef);
        previousRoleDef = newRoleDef;
        return this;
    }

    public RbacPermissionDefinition createPermission(final Permission permission) {
        return createPermission(rootEntityAlias, permission);
    }

    private RbacPermissionDefinition createPermission(final EntityAlias entityAlias, final Permission permission) {
        final RbacPermissionDefinition permDef = new RbacPermissionDefinition(entityAlias, permission, true);
        permDefs.add(permDef);
        return permDef;
    }

    public <EC extends RbacObject> RbacView declarePlaceholderEntityAliases(final String... aliasNames) {
        for ( String alias: aliasNames ) {
            entityAliases.put(alias, new EntityAlias(alias));
        }
        return this;
    }

    public <EC extends RbacObject> RbacView importRootEntityAliasProxy(
            final String aliasName, final Class entityClass, final SQL fetchSql, final Column dependsOnColum) {
        if ( rootEntityAliasProxy != null ) {
            throw new IllegalStateException("there is already an entityAliasProxy: " + rootEntityAliasProxy);
        }
        rootEntityAliasProxy = importEntityAliasImpl(aliasName, entityClass, fetchSql, dependsOnColum);
        return this;
    }

    public RbacView importEntityAlias(
            final String aliasName, final Class entityClass, final SQL fetchSql, final Column dependsOnColum) {
        importEntityAliasImpl(aliasName, entityClass, fetchSql, dependsOnColum);
        return this;
    }

    private EntityAlias importEntityAliasImpl(final String aliasName, final Class entityClass, final SQL fetchSql, final Column dependsOnColum) {
        final var entityAlias = new EntityAlias(aliasName, entityClass, fetchSql, dependsOnColum);
        entityAliases.put(aliasName, entityAlias);
        try {
            importAsAlias(aliasName, rbacDefinition(entityClass));
        } catch ( final Exception exc) {
            new RuntimeException("cannot import entity: " + entityClass, exc);
        }
        return entityAlias;
    }

    private static RbacView rbacDefinition(final Class<? extends RbacObject> entityClass)
            throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
        return (RbacView) entityClass.getMethod("rbac").invoke(null);
    }

    private RbacView importAsAlias(final String aliasName, final RbacView importedRbacView) {
        final var mapper = new AliasNameMapper(importedRbacView, aliasName);
        importedRbacView.getEntityAliases().values().stream()
                .filter(entityAlias -> !importedRbacView.isRootEntityAlias(entityAlias))
                .filter(entityAlias -> !entityAlias.isGlobal())
                .forEach(entityAlias -> {
                    final String mappedAliasName = mapper.map(entityAlias.aliasName);
                    entityAliases.put(mappedAliasName, new EntityAlias(mappedAliasName, entityAlias.entityClass));
                });
        importedRbacView.getRoleDefs().forEach(roleDef -> {
            new RbacRoleDefinition( findEntityAlias(mapper.map(roleDef.entityAlias.aliasName)), roleDef.role);
        });
        importedRbacView.getGrantDefs().forEach(grantDef -> {
            if (grantDef.grantType() == RbacGrantDefinition.GrantType.ROLE_TO_ROLE) {
                findOrCreateGrantDef(
                        findRbacRole(mapper.map(grantDef.getSubRoleDef().entityAlias.aliasName), grantDef.getSubRoleDef().getRole()),
                        findRbacRole(mapper.map(grantDef.getSuperRoleDef().entityAlias.aliasName), grantDef.getSuperRoleDef().getRole())
                );
            }
        });
        return this;
    }

    public RbacGrantBuilder toRole(final String entityAlias, final Role role) {
        return new RbacGrantBuilder(entityAlias, role);
    }

    public RbacExampleRole forExampleRole(final String entityAlias, final Role role) {
        return new RbacExampleRole(entityAlias, role);
    }


    private  RbacGrantDefinition grantRoleToUser(final RbacRoleDefinition roleDefinition, final RbacUserReference user) {
        return findOrCreateGrantDef(roleDefinition, user).toCreate();
    }

    private  RbacGrantDefinition grantPermissionToRole(final RbacPermissionDefinition permDef , final RbacRoleDefinition roleDef) {
        return findOrCreateGrantDef(permDef, roleDef).toCreate();
    }

    private  RbacGrantDefinition grantSubRoleToSuperRole(final RbacRoleDefinition subRoleDefinition, final RbacRoleDefinition superRoleDefinition) {
        return findOrCreateGrantDef(subRoleDefinition, superRoleDefinition).toCreate();
    }

    boolean isRootEntityAlias(final EntityAlias entityAlias) {
        return entityAlias == this.rootEntityAlias;
    }

    public boolean isEntityAliasProxy(final EntityAlias entityAlias) {
        return entityAlias == rootEntityAliasProxy;
    }

    public class RbacGrantBuilder {

        private final RbacRoleDefinition superRoleDef;

        private RbacGrantBuilder(final String entityAlias, final Role role) {
            this.superRoleDef = findRbacRole(entityAlias, role);
        }

        public RbacView grantRole(final String entityAlias, final Role role) {
            findOrCreateGrantDef(findRbacRole(entityAlias, role), superRoleDef).toCreate();
            return RbacView.this;
        }

    }

    @Getter
    @EqualsAndHashCode
    public class RbacGrantDefinition {
        private final RbacUserReference userDef;
        private final RbacRoleDefinition superRoleDef;
        private final RbacRoleDefinition subRoleDef;
        private final RbacPermissionDefinition permDef;
        private boolean toCreate;

        @Override
        public String toString() {
            final var arrow = isAssumed() ? " --> " : " -- // --> ";
            return switch (grantType()) {
                case ROLE_TO_USER -> userDef.toString() + arrow + subRoleDef.toString();
                case ROLE_TO_ROLE -> superRoleDef + arrow + subRoleDef;
                case PERM_TO_ROLE -> superRoleDef + arrow + permDef;
            };
        }

        RbacGrantDefinition(final RbacRoleDefinition subRoleDef, final RbacRoleDefinition superRoleDef) {
            this.userDef = null;
            this.subRoleDef = subRoleDef;
            this.superRoleDef = superRoleDef;
            this.permDef = null;
            register(this);
        }

        public RbacGrantDefinition(final RbacPermissionDefinition permDef, final RbacRoleDefinition roleDef) {
            this.userDef = null;
            this.subRoleDef = null;
            this.superRoleDef = roleDef;
            this.permDef = permDef;
            register(this);
        }

        public RbacGrantDefinition(final RbacRoleDefinition roleDef, final RbacUserReference userDef) {
            this.userDef = userDef;
            this.subRoleDef = roleDef;
            this.superRoleDef = null;
            this.permDef = null;
            register(this);
        }

        private void register(final RbacGrantDefinition rbacGrantDefinition) {
            grantDefs.add(rbacGrantDefinition);
        }

        @NotNull
        GrantType grantType() {
            return permDef != null ? GrantType.PERM_TO_ROLE
                    : userDef != null ? GrantType.ROLE_TO_USER
                    : GrantType.ROLE_TO_ROLE;
        }

        boolean isAssumed() {
            // TODO: not implemented yet
            return true;
        }

        boolean isToCreate() {
            return toCreate;
        }

        RbacGrantDefinition toCreate() {
            toCreate = true;
            return this;
        }

        public enum GrantType {
            ROLE_TO_USER,
            ROLE_TO_ROLE,
            PERM_TO_ROLE
        }
    }

    public class RbacExampleRole {

        final EntityAlias subRoleEntity;
        final Role subRole;
        private EntityAlias superRoleEntity;
        Role superRole;

        public RbacExampleRole(final String entityAlias, final Role role) {
            this.subRoleEntity = findEntityAlias(entityAlias);
            this.subRole = role;
        }

        public RbacView wouldBeGrantedTo(final String entityAlias, final Role role) {
            this.superRoleEntity = findEntityAlias(entityAlias);
            this.superRole = role;
            return RbacView.this;
        }
    }

    @Getter
    @EqualsAndHashCode
    public class RbacPermissionDefinition {

        final EntityAlias entityAlias;
        final Permission permission;
        final boolean toCreate;

        public RbacPermissionDefinition(final EntityAlias entityAlias, final Permission permission, final boolean toCreate) {
            this.entityAlias = entityAlias;
            this.permission = permission;
            this.toCreate = toCreate;
        }

        public RbacView pop() {
            return RbacView.this;
        }

        public RbacPermissionDefinition withIncomingSuperRole(
                final Class<HsOfficeRelationshipEntity> hsOfficeRelationshipEntityClass,
                final Role owner) {

            return this;
        }

        public RbacPermissionDefinition  grantedTo(final String entityAlias, final Role role) {
            findOrCreateGrantDef(this, findRbacRole(entityAlias, role) ).toCreate();
            return this;
        }

        @Override
        public String toString() {
            return "perm:" + entityAlias.aliasName + permission;
        }
    }

    @Getter
    @EqualsAndHashCode
    public class RbacRoleDefinition {

        private final EntityAlias entityAlias;
        private final Role role;
        private boolean toCreate;

        public RbacRoleDefinition(final EntityAlias entityAlias, final Role role) {
            this.entityAlias = entityAlias;
            this.role = role;
            roleDefs.add(this);
        }

        public RbacRoleDefinition toCreate() {
            this.toCreate = true;
            return this;
        }

        public RbacRoleDefinition owningUser(final RbacUserReference.UserRole userRole) {
            grantRoleToUser(this, findUserRef(userRole));
            return this;
        }

        public RbacRoleDefinition permission(final Permission permission) {
            grantPermissionToRole( createPermission(entityAlias, permission) , this);
            return this;
        }

        public RbacRoleDefinition incomingSuperRole(final String entityAlias, final Role role) {
            final var incomingSuperRole =  findRbacRole(entityAlias, role);
            grantSubRoleToSuperRole(this, incomingSuperRole);
            return this;
        }

        public RbacRoleDefinition outgoingSubRole(final String entityAlias, final Role role) {
            final var outgoingSubRole =  findRbacRole(entityAlias, role);
            grantSubRoleToSuperRole(outgoingSubRole, this);
            return this;
        }

        @Override
        public String toString() {
            return "role:" + entityAlias.aliasName + role;
        }
    }

    public RbacUserReference findUserRef(final RbacUserReference.UserRole userRole) {
        return userDefs.stream().filter(u -> u.role == userRole).findFirst().orElseThrow();
    }

    @EqualsAndHashCode
    public class RbacUserReference {

        public enum UserRole {
            CREATOR
        }

        final UserRole role;
        public RbacUserReference(final UserRole creator) {
            this.role = creator;
            userDefs.add(this);
        }

        @Override
        public String toString() {
            return "user:" + role;
        }
    }

    EntityAlias findEntityAlias(final String aliasName) {
        final var found = entityAliases.get(aliasName);
        if ( found == null ) {
            throw new IllegalArgumentException("entityAlias not found: " + aliasName);
        }
        return found;
    }

    RbacRoleDefinition findRbacRole(final EntityAlias entityAlias, final Role role) {
        return roleDefs.stream()
                .filter(r -> r.getEntityAlias() == entityAlias && r.getRole().equals(role))
                .findFirst()
                .orElseGet(() -> new RbacRoleDefinition(entityAlias, role));
    }

    public RbacRoleDefinition findRbacRole(final String entityAliasName, final Role role) {
        return findRbacRole(findEntityAlias(entityAliasName), role);

    }

    private RbacGrantDefinition findOrCreateGrantDef(final RbacRoleDefinition roleDefinition, final RbacUserReference user) {
        return grantDefs.stream()
                .filter(g -> g.subRoleDef == roleDefinition && g.userDef == user)
                .findFirst()
                .orElseGet(() -> new RbacGrantDefinition(roleDefinition, user));
    }

    private RbacGrantDefinition findOrCreateGrantDef(final RbacPermissionDefinition permDef, final RbacRoleDefinition roleDef) {
        return grantDefs.stream()
                .filter(g -> g.permDef == permDef && g.subRoleDef == roleDef)
                .findFirst()
                .orElseGet(() -> new RbacGrantDefinition(permDef, roleDef));
    }

    private RbacGrantDefinition findOrCreateGrantDef(final RbacRoleDefinition subRoleDefinition, final RbacRoleDefinition superRoleDefinition) {
        return grantDefs.stream()
                .filter(g -> g.subRoleDef == subRoleDefinition && g.superRoleDef == superRoleDefinition)
                .findFirst()
                .orElseGet(() -> new RbacGrantDefinition(subRoleDefinition, superRoleDefinition));
    }

    record EntityAlias(String aliasName, Class<? extends RbacObject> entityClass, SQL fetchSql, Column dependsOnColum) {

            public EntityAlias(final String aliasName) {
                this(aliasName, null, null, null);
            }

            public EntityAlias(final String aliasName, final Class<? extends RbacObject> entityClass) {
                this(aliasName, entityClass, null, null);
            }

        boolean isGlobal() {
            return aliasName().equals("global");
        }

        boolean isPlaceholder() {
                return entityClass == null;
        }



        private String withoutEntitySuffix(final String simpleEntityName) {
            return simpleEntityName.substring(0, simpleEntityName.length()-"Entity".length());
        }

        String simpleName() {
            return isGlobal()
                    ? aliasName
                    : uncapitalize(withoutEntitySuffix(entityClass.getSimpleName()));
        }
    }

    public record Role(String roleName) {
        public static final Role OWNER = new Role("owner");
        public static final Role ADMIN = new Role("admin");
        public static final Role AGENT = new Role("agent");
        public static final Role TENANT = new Role("tenant");
        public static final Role REFERRER = new Role("referrer");

        @Override
        public String toString() {
            return ":" + roleName;
        }

        @Override
        public boolean equals(final Object obj) {
            return ((obj instanceof Role) && ((Role)obj).roleName.equals(this.roleName));
        }
    }

    public record Permission(String permission) {
        public static final Permission ALL = new Permission("*");
        public static final Permission EDIT = new Permission("edit");
        public static final Permission VIEW = new Permission("view");

        public static Permission custom(final String permission) {
            return new Permission(permission);
        }

        @Override
        public String toString() {
            return ":" + permission;
        }
    }

    public static class SQL {

        public static SQL fetchedBySql(final String sql) {
            return new SQL(sql);
        }

        public static SQL query(final String sql) {
            if (sql.matches(";[ \t]*$")) {
                throw new IllegalArgumentException("SQL expression must not end with ';': " + sql);
            }
            return new SQL(sql);
        }

        public final String sql;

        private SQL(final String sql) {
            this.sql = sql;
        }
    }

    public static class Column {

        public static Column dependsOnColumn(final String column) {
            return new Column(column);
        }

        public final String column;

        private Column(final String column) {
            this.column = column;
        }
    }

    private static class AliasNameMapper {

        private final RbacView importedRbacView;
        private final String outerAliasName;

        AliasNameMapper(final RbacView importedRbacView, final String outerAliasName) {
            this.importedRbacView = importedRbacView;
            this.outerAliasName = outerAliasName;
        }

        String map(final String originalAliasName) {
            if (originalAliasName.equals(importedRbacView.rootEntityAlias.aliasName) ) {
                return outerAliasName;
            }
            if (originalAliasName.equals("global") ) {
                return originalAliasName;
            }
            return outerAliasName + "." + originalAliasName;
        }
    }
}
