package net.hostsharing.hsadminng.rbac.rbacdef;

import net.hostsharing.hsadminng.rbac.rbacdef.RbacView.RbacPermissionDefinition;

import jakarta.persistence.Table;
import java.util.HashSet;
import java.util.Set;

import static java.util.stream.Collectors.*;
import static net.hostsharing.hsadminng.rbac.rbacdef.PostgresTriggerReference.NEW;
import static net.hostsharing.hsadminng.rbac.rbacdef.RbacView.RbacGrantDefinition.GrantType.*;
import static net.hostsharing.hsadminng.rbac.rbacdef.RbacView.Role.*;
import static org.apache.commons.lang3.StringUtils.capitalize;
import static org.apache.commons.lang3.StringUtils.uncapitalize;

class RolesGrantsAndPermissionsGenerator {

    private final RbacView rbacDef;
    private final Set<RbacView.RbacGrantDefinition> rbacGrants = new HashSet<>();
    private final String liquibaseTagPrefix;
    private final Class<?> entityClass;
    private final String simpleEntityName;
    private final String simpleEntityVarName;
    private final String rawTableName;

    RolesGrantsAndPermissionsGenerator(final RbacView rbacDef, final String liquibaseTagPrefix) {
        this.rbacDef = rbacDef;
        this.rbacGrants.addAll(rbacDef.getGrantDefs().stream()
                .filter(RbacView.RbacGrantDefinition::isToCreate)
                .collect(toSet()));
        this.liquibaseTagPrefix = liquibaseTagPrefix;
        
        entityClass = rbacDef.getRootEntityAlias().entityClass();
        simpleEntityVarName = rbacDef.getRootEntityAlias().simpleName();
        simpleEntityName = capitalize(simpleEntityVarName);
        rawTableName = getRawTableName(entityClass);
    }

    void generateTo(final StringWriter plPgSql) {
        generateHeader(plPgSql);
        generateTriggerFunction(plPgSql);
        generageInsertTrigger(plPgSql);
        generateFooter(plPgSql);
    }

    private void generateHeader(final StringWriter plPgSql) {
        plPgSql.writeLn("""
                -- ============================================================================
                --changeset ${liquibaseTagPrefix}-rbac-CREATE-ROLES-GRANTS-PERMISSIONS:1 endDelimiter:--//
                -- ----------------------------------------------------------------------------
                """
                .replace("${liquibaseTagPrefix}", liquibaseTagPrefix));
    }

    private void generateTriggerFunction(final StringWriter plPgSql) {
        plPgSql.writeLn("""
                /*
                    Creates the roles, grants and permission for the AFTER INSERT TRIGGER.
                 */
                                
                create or replace function createRbacRolesFor${simpleEntityName}()
                    returns trigger
                    language plpgsql
                    strict as $$
                declare
                """
                .replace("${simpleEntityName}", simpleEntityName));

        plPgSql.indented(() -> {
            rbacDef.getEntityAliases().values().stream()
                    .filter((ea) -> !rbacDef.isRootEntityAlias(ea))
                    .filter((ea) -> ea.fetchSql() != null)
                    .forEach((ea) -> {
                        plPgSql.writeLn( entityRefVar(NEW, ea) + " " + getRawTableName(ea.entityClass()) + ";");
                    });
            });

        plPgSql.writeLn("""
                begin
                    if TG_OP <> 'INSERT' then
                        raise exception 'invalid usage of TRIGGER AFTER INSERT function';
                    end if;
                """);

        plPgSql.indented(() -> {

            rbacDef.getEntityAliases().values().stream()
                    .filter((ea) -> !rbacDef.isRootEntityAlias(ea))
                    .filter((ea) -> ea.fetchSql() != null)
                    .forEach((ea) -> {
                plPgSql.writeLn( ea.fetchSql().sql.replace("${ref}", NEW.name())  + " into " + entityRefVar(NEW, ea) + ";");
            });

            createRolesWithGrantsSql(plPgSql, OWNER);
            createRolesWithGrantsSql(plPgSql, ADMIN);
            createRolesWithGrantsSql(plPgSql, AGENT);
            createRolesWithGrantsSql(plPgSql, TENANT);
            createRolesWithGrantsSql(plPgSql, REFERRER);

            plPgSql.writeLn();
            rbacGrants.forEach(g -> plPgSql.writeLn(generateGrant(g)));

            plPgSql.writeLn("return NEW;");
        });

        plPgSql.writeLn("end; $$;");
        plPgSql.writeLn();
    }

    private String generateGrant(RbacView.RbacGrantDefinition grantDef) {
        return switch (grantDef.grantType()) {
            case ROLE_TO_USER -> throw new IllegalArgumentException("unexpected grant");
            case ROLE_TO_ROLE -> "call grantRoleToRole(${subRoleRef}, ${superRoleRef}));"
                        .replace("${subRoleRef}", roleRef(NEW, grantDef.getSubRoleDef()) )
                        .replace("${superRoleRef}", roleRef(NEW, grantDef.getSuperRoleDef()) );
            case PERM_TO_ROLE -> "call grantPermissionsToRole(${permRef}, ${superRoleRef}));"
                    .replace("${permRef}", permRef(NEW, grantDef.getPermDef()) )
                    .replace("${superRoleRef}", roleRef(NEW, grantDef.getSuperRoleDef()) );
        };
    }

    private String permRef(final PostgresTriggerReference ref, final RbacPermissionDefinition permDef) {
        return "createPermissions(${entityRef}.uuid, array ['${perm}']"
                .replace("${entityRef}", rbacDef.isRootEntityAlias(permDef.entityAlias)
                        ? ref.name()
                        : "not implemented yet" ) // TODO
                .replace("${perm}", permDef.permission.permission());
    }

    private String getRawTableName(final Class<?> entityClass) {
        return withoutRvSuffix(entityClass.getAnnotation(Table.class).name());
    }

    private String roleRef(final PostgresTriggerReference rootRefVar, final RbacView.RbacRoleDefinition roleDef) {
        if ( roleDef == null ) {
            System.out.println("null");
        }
        if ( roleDef.getEntityAlias().isGlobal()) {
            return "globalAdmin()";
        }
        final String entityRefVar = entityRefVar(rootRefVar, roleDef.getEntityAlias());
        return roleDef.getEntityAlias().simpleName() + capitalize(roleDef.getRole().roleName())
                + "(" + entityRefVar + ")";
    }

    private static String entityRefVar(
            final PostgresTriggerReference rootRefVar,
            final RbacView.EntityAlias entityAlias) {
        return rootRefVar.name().toLowerCase() + capitalize(entityAlias.aliasName());
    }

    private void createRolesWithGrantsSql(final StringWriter plPgSql, final RbacView.Role role) {

         final var isToCreate = rbacDef.getRoleDefs().stream()
                 .filter(roleDef -> rbacDef.isRootEntityAlias(roleDef.getEntityAlias()) && roleDef.getRole() == role )
                 .findFirst().map(RbacView.RbacRoleDefinition::isToCreate).orElse(false);
        if (!isToCreate) {
            return;
        }

        plPgSql.writeLn();
        plPgSql.writeLn("perform createRoleWithGrants(");
        plPgSql.indented( () -> {
            plPgSql.writeLn("${simpleVarName)${roleSuffix}(NEW),"
                    .replace("${simpleVarName)", simpleEntityVarName)
                    .replace("${roleSuffix}", capitalize(role.roleName())));

            final var permissionGrantsForRole = findPermissionsGrantsForRole(rbacDef.getRootEntityAlias(), role);
            if (!permissionGrantsForRole.isEmpty()) {
                final var permissionsForRoleInPlPgSql = permissionGrantsForRole.stream()
                        .map(RbacView.RbacGrantDefinition::getPermDef)
                        .map(RbacPermissionDefinition::getPermission)
                        .map(RbacView.Permission::permission)
                        .map(p -> "'" + p + "'")
                        .collect(joining(", "));
                plPgSql.indented( () ->
                        plPgSql.writeLn("permissions => array[" + permissionsForRoleInPlPgSql + "],\n"));
                rbacGrants.removeAll(permissionGrantsForRole);
            }

            final var grantsToUsers = findGrantsToUserForRole(rbacDef.getRootEntityAlias(), role);
            if (!grantsToUsers.isEmpty()) {
                final var grantsToUsersPlPgSql = grantsToUsers.stream()
                        .map(RbacView.RbacGrantDefinition::getUserDef)
                        .map(this::toPlPgSqlReference)
                        .collect(joining(", "));
                plPgSql.indented(() ->
                        plPgSql.writeLn("userUuids => array[" + grantsToUsersPlPgSql + "],\n"));
                rbacGrants.removeAll(grantsToUsers);
            }

            final var incomingGrants = findIncomingSuperRolesForRole(rbacDef.getRootEntityAlias(), role);
            if (!incomingGrants.isEmpty()) {
                final var incomingGrantsInPlPgSql = incomingGrants.stream()
                        .map(RbacView.RbacGrantDefinition::getSuperRoleDef)
                        .map(r -> toPlPgSqlReference(NEW, r))
                        .collect(joining(", "));
                plPgSql.indented(() ->
                        plPgSql.writeLn("incomingSuperRoles => array[" + incomingGrantsInPlPgSql + "],\n"));
                rbacGrants.removeAll(incomingGrants);
            }

            final var outgoingGrants = findOutgoingSuperRolesForRole(rbacDef.getRootEntityAlias(), role);
            if (!outgoingGrants.isEmpty()) {
                final var outgoingGrantsInPlPgSql = outgoingGrants.stream()
                        .map(RbacView.RbacGrantDefinition::getSuperRoleDef)
                        .map(r -> toPlPgSqlReference(NEW, r))
                        .collect(joining(", "));
                plPgSql.indented(() ->
                        plPgSql.writeLn("outgoingSubRoles => array[" + outgoingGrantsInPlPgSql + "],\n"));
                rbacGrants.removeAll(outgoingGrants);
            }

            plPgSql.chopTail(",\n");
            plPgSql.writeLn();
        });

        plPgSql.writeLn(");");
    }

    private Set<RbacView.RbacGrantDefinition> findPermissionsGrantsForRole(final RbacView.EntityAlias entityAlias, final RbacView.Role role) {
        final var roleDef = rbacDef.findRbacRole(entityAlias, role);
        return rbacGrants.stream()
                .filter(g -> g.grantType() == PERM_TO_ROLE && g.getSuperRoleDef()==roleDef )
                .collect(toSet());
    }

    private Set<RbacView.RbacGrantDefinition> findGrantsToUserForRole(final RbacView.EntityAlias entityAlias, final RbacView.Role role) {
        final var roleDef = rbacDef.findRbacRole(entityAlias, role);
        return rbacGrants.stream()
                .filter(g -> g.grantType() == ROLE_TO_USER && g.getSubRoleDef() == roleDef )
                .collect(toSet());
    }

    private Set<RbacView.RbacGrantDefinition> findIncomingSuperRolesForRole(final RbacView.EntityAlias entityAlias, final RbacView.Role role) {
        final var roleDef = rbacDef.findRbacRole(entityAlias, role);
        return rbacGrants.stream()
                .filter(g -> g.grantType() == ROLE_TO_ROLE && g.getSubRoleDef()==roleDef )
                .collect(toSet());
    }

    private Set<RbacView.RbacGrantDefinition> findOutgoingSuperRolesForRole(final RbacView.EntityAlias entityAlias, final RbacView.Role role) {
        final var roleDef = rbacDef.findRbacRole(entityAlias, role);
        return rbacGrants.stream()
                .filter(g -> g.grantType() == ROLE_TO_ROLE && g.getSuperRoleDef()==roleDef )
                .filter(g -> g.getSubRoleDef().getEntityAlias() != entityAlias)
                .collect(toSet());
    }

    private void generageInsertTrigger(final StringWriter plPgSql) {
        plPgSql.writeLn("""
                /*
                    An AFTER INSERT TRIGGER which creates the role structure for a new ${simpleEntityName}
                 */
                                
                create trigger createRbacRolesFor${simpleEntityName}_Trigger
                    after insert
                    on ${rawTableName}
                    for each row
                execute procedure createRbacRolesFor${simpleEntityName}();
                --//
                """
                .replace("${simpleEntityName}", simpleEntityName)
                .replace("${rawTableName}", rawTableName)
        );
    }

    private static void generateFooter(final StringWriter plPgSql) {
        plPgSql.writeLn();
        plPgSql.writeLn();
    }

    private String withoutRvSuffix(final String tableName) {
        return tableName.substring(0, tableName.length()-"_rv".length());
    }

    private String toPlPgSqlReference(final RbacView.RbacUserReference userRef) {
        return switch (userRef.role) {
            case CREATOR -> "currentUserUuid()";
            default -> throw new IllegalArgumentException("unknown user role: " + userRef);
        };
    }

    private String toPlPgSqlReference(final PostgresTriggerReference triggerRef, final RbacView.RbacRoleDefinition roleDef) {
        return toVar(roleDef) +
                (roleDef.getEntityAlias().isGlobal() ? "()"
                        : rbacDef.isRootEntityAlias(roleDef.getEntityAlias()) ? ("(" + triggerRef.name() + ")")
                        : "(" + toTriggerReference(triggerRef, roleDef.getEntityAlias()) + ")");
    }

    private static String toVar(final RbacView.RbacRoleDefinition roleDef) {
        return uncapitalize(roleDef.getEntityAlias().simpleName()) + capitalize(roleDef.getRole().roleName());
    }

    private static String toTriggerReference(final PostgresTriggerReference triggerRef, final RbacView.EntityAlias entityAlias) {
        return triggerRef.name().toLowerCase() + capitalize(entityAlias.aliasName());
    }
}
